/* lcs.c *****************************************************\
*                                                             *
*  First order Godunov-type 1D scheme with                    *
*  Linearized Characteristic-Based Appr. Riemann Solver (LCS) *
*  written by  Andrei A. Chernousov                           *
*  E-mail: <andrei99@iname.com>                               *
*  http://www.geocities.com/andrei_chernousov                 *
*  November 12, 2001                                          *
*                                                             *
\*************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#define K     1.4
#define K_1_2 0.2
#define K_1   0.4
#define _05K  0.7
#define _025K 0.35
#define K_K_1 3.5

float
  *U1, *U2, *U3, /* conserved quantities */
  *Jp, *Jm, *Jo, /* 'invariants' of the linearized system */
  *F1, *F2, *F3, /* fluxes */

  J_p,  J_m,  J_o,
  al_p, al_m, al_o,

  _jp, _jm, _jo,
  jp_,  jm_,  jo_,

/* C_p_, C_m_, C_o_, */
  C_p, C_m, C_o,

  P, R, U, C, PRC, /* primitive variables */

  deltaX, deltaT, deltaT_X;

unsigned LEN, LENN, i, i_, step, numstep;

FILE *pF;

/************
*  Array1D  *
************/
float *Array1D(unsigned size)
{
  float *x;

  if ((x = (float*)malloc(size * sizeof(float))) == NULL) {
    fprintf(stderr, "lcs: can't allocate memory\n");
    exit(-1);
  }
   
  return x;
   
} /* end Array1D() */


/*************
* Initialize *
*************/
void Initialize(void)
{
  char str[80];

  /* loading data from file "lcs.ini" */
  if ((pF = fopen("lcs.ini", "r")) == NULL) {
    fprintf(stderr, "lcs: can't open file \"lcs.ini\"\n");
    exit(-1);
  }
  
  i = 0;
  do {
    if (fgets(str, 80, pF) != NULL) {
      switch (i) {
      case 0:  LEN    = atoi(str); break;
      case 1:  deltaX = atof(str); break;
      case 2:  deltaT = atof(str); break;
      case 3:  numstep= atoi(str); break;
      default: break;
      } /* end switch() */
      i++;
    } /* end if() */
    else {
      fprintf(stderr, "lcs: error reading file \"lcs.ini\"\n");
      exit(-1);
    }
  } while (i < 4);
  fclose(pF);

  deltaT_X = deltaT / deltaX;
  LENN = LEN+1;

  /* dynamic memory allocation */
  U1 = Array1D(LEN+2); U2 = Array1D(LEN+2); U3 = Array1D(LEN+2);
  Jp = Array1D(LEN+2); Jm = Array1D(LEN+2); Jo = Array1D(LEN+2);
  F1 = Array1D(LEN+1); F2 = Array1D(LEN+1); F3 = Array1D(LEN+1);

  /* initial conditions */
  for (i = 1, U = 0., R = 2., P = 200000.; i <= LEN/2; i++) {
    U1[i] = R;
    U2[i] = R * U;
    U3[i] = P / K_1 + 0.5 * R * U * U;
    C = sqrt(K * P / R);
    Jo[i] = P - C * C * R;
    Jp[i] = U + (PRC = P / (R * C));
    Jm[i] = U - PRC;
  } /* end for() */
  for (U = 0., R = 1., P = 100000.; i < LENN; i++) {
    U1[i] = R;
    U2[i] = R * U;
    U3[i] = P / K_1 + 0.5 * R * U * U;
    C = sqrt(K * P / R);
    Jo[i] = P - C * C * R;
    Jp[i] = U + (PRC = P / (R * C));
    Jm[i] = U - PRC;
  }

} /* end Initialize() */


/***********************
* BounCondInGhostCells *
***********************/
void BounCondInGhostCells(void)
{
  /* non-reflective? */
  Jp[0] = Jp[1];
  Jm[0] = Jm[1];
  Jo[0] = Jo[1];
  Jp[LEN+1] = Jp[LEN];
  Jm[LEN+1] = Jm[LEN];
  Jo[LEN+1] = Jo[LEN];
  /* solid walls
     Jp[0] = - Jm[1];
     Jm[0] = - Jp[1];
     Jo[0] = Jo[1];
     Jp[LEN+1] = - Jm[LEN];
     Jm[LEN+1] = - Jp[LEN];
     Jo[LEN+1] = Jo[LEN];
  */
} /* end BounCondInGhostCells() */


/***********
*  Fluxes  *   LCS: the fastest possible Approximate Riemann Solver ?!
***********/
void Fluxes(void)
{
  for (i = 0, i_ = 1; i < LENN; i++, i_++) {
    /* 'invariants' to the left */
    _jp = Jp[i ], _jm = Jm[i ], _jo = Jo[i ];
    /* 'invariants' to the right */
    jp_ = Jp[i_], jm_ = Jm[i_], jo_ = Jo[i_];
    /* initial guess for U, C at cell interface */
    U = 0.25  * (_jp + _jm + jp_ + jm_);
    C = _025K * (_jp - _jm + jp_ - jm_);
    /* initial guess for characteristics speeds */
    C_p = U + C;
    C_m = U - C;
    C_o = U;
    /* essentially noniterative solver */
    /*  proc: */
    /* J_p and al_p */
    if (C_p > 0.) { al_p = K_1_2 * (_jm - _jp) / _jo; J_p  = _jp; }
    else          { al_p = K_1_2 * (jm_ - jp_) / jo_; J_p  = jp_; }
    /* J_m and al_m */
    if (C_m > 0.) { al_m = K_1_2 * (_jp - _jm) / _jo; J_m  = _jm; }
    else          { al_m = K_1_2 * (jp_ - jm_) / jo_; J_m  = jm_; }
    /* J_o and al_o */
    if (C_o > 0.) { 
      al_o = _05K * (_jp - _jm); 
      al_o = - al_o * al_o; 
      J_o  = _jo;
    }
    else { 
      al_o = _05K * (jp_ - jm_); 
      al_o = - al_o * al_o;
      J_o  = jo_;
    }
    /* parameters on the interface */
    P = (J_p - J_m) / (al_p - al_m);
    U = J_p - al_p * P;
    R = (J_o - P) / al_o;
    /* check signs of the Ch-velocities (optional) */
    /*	  C = sqrt( K * P / R );
	  C_p_ = U + C; C_m_ = U - C; C_o_ = U;
	  if( C_p_*C_p < 0 || C_m_*C_m < 0 || C_o_*C_o < 0 ) {
	  C_p = C_p_; C_m = C_m_; C_o = C_o_;
	  goto proc;
	  }
    */	     
    /* fluxes */
    F1[i] = R * U;
    F2[i] = R * U * U + P;
    F3[i] = U * (K_K_1 * P + 0.5 * R * U * U);
  }

} /* end Fluxes() */


/**************
*  Evolution  *
**************/
void Evolution(void)
{
  for (i = 1; i < LENN; i++) {
    U1[i] += deltaT_X * (F1[i-1] - F1[i]);
    U2[i] += deltaT_X * (F2[i-1] - F2[i]);
    U3[i] += deltaT_X * (F3[i-1] - F3[i]);
    R =  U1[i];
    U =  U2[i] / R;
    P = (U3[i] - 0.5 * R * U * U) * K_1;
    C = sqrt(K * P / R);
    Jo[i] = P - C * C * R;
    Jp[i] = U + ( PRC = P / ( R * C ) );
    Jm[i] = U - PRC;
  }
} /* end Evolution() */


/***********
*  Output  *
***********/
void Output(void)
{
  /* output to "output.dat" file */
  if ((pF = fopen("output.dat", "w")) == NULL) {
    fprintf(stderr, "lcs: can't open \"output.dat\"\n");
    exit(-1);
  }
/*  fprintf(pF, "%d\n%d\n", LEN+2, 3);
*/
  fprintf(pF, "0. 90000. 0.9 0. \n");
  for (i = 1; i < LENN; i++) {
    R =  U1[i];
    U =  U2[i] / R;
    P = (U3[i] - 0.5 * R * U * U) * K_1;
    fprintf( pF,"%3.3f %6.0f %1.4f %3.1f \n", (i - 0.5) * deltaX, P, R, U);
  }
  fprintf(pF,"%d 210000. 2.1 100. \n", LEN );
  fclose(pF);
 
} /* end Output() */


/*********
*  MAIN  *
**********/
int main(void)
{
  /*--- initialization ---*/
  Initialize();
   
  /*-- time stepping ---*/
  while (step < numstep) {
    /*--- print out step number ---*/
    //printf("%d\n", ++step);
    step++;
    /*--- bc in ghost cells ---*/
    BounCondInGhostCells();
    /*--- fluxes at cell interfaces ---*/
    Fluxes();
    /*--- evolution ---*/
    Evolution();
  } /* end while */
   
  /*--- output of the flowfield ---*/
  Output();
   
  return 0;
   
} /* end main() */

/*--- end lcs.c ---*/
